/*
 * Copyright (C) 2022 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *      Author: Suvendhu Hansa <suvendhu.hansa@iopsys.eu>
 */

#include <libbbfdm_api.h>
#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE
#endif

#include <unistd.h>
#include <limits.h>
#include <pwd.h>
#include <shadow.h>

#include "users.h"

enum group_operations {
	NAME_CHANGE,
	DELETE,
	__OP_MAX
};

typedef enum {
	Credentials_Good,
	Credentials_Bad_Requested_Username_Not_Supported,
	Credentials_Bad_Requested_Password_Incorrect,
	Credentials_Missing,
	Error_Invalid_Input,
	Error_Other
} e_cred_t;

e_cred_t check_user_cred(char *username, char *password)
{
	char *encrypted;
	struct passwd *pwd;
	struct spwd *spwd;

	if ((DM_STRLEN(username) & DM_STRLEN(password)) == 0) {
		return Error_Invalid_Input;
	}

	/* Look up password and shadow password records for username */
	pwd = getpwnam(username);
	if (pwd == NULL) {
		return Credentials_Bad_Requested_Username_Not_Supported;
	}

	spwd = getspnam(username);
	if (spwd == NULL && errno == EACCES) {
		return Error_Other;
	}

	/* Use the shadow password */
	if (spwd != NULL) {
		pwd->pw_passwd = spwd->sp_pwdp;
	}

	encrypted = crypt(password, pwd->pw_passwd);
	if (encrypted == NULL) {
		return Error_Other;
	}

	if (strcmp(encrypted, pwd->pw_passwd)) {
		return Credentials_Bad_Requested_Password_Incorrect;
	}

	return Credentials_Good;
}


static operation_args checkCred_args = {
	.in = (const char *[]) {
		"Username",
		"Password",
		NULL
	},
	.out = (const char *[]) {
		"Status",
		NULL
	}
};

static int get_operate_args_CheckCred(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = (char *)&checkCred_args;
	return 0;
}

static char *get_cred_result(e_cred_t ret)
{
	switch(ret) {
		case Credentials_Good:
			return "Credentials_Good";
		case Credentials_Bad_Requested_Username_Not_Supported:
			return "Credentials_Bad_Requested_Username_Not_Supported";
		case Credentials_Bad_Requested_Password_Incorrect:
			return "Credentials_Bad_Requested_Password_Incorrect";
		case Credentials_Missing:
			return "Credentials_Missing";
		case Error_Invalid_Input:
			return "Error_Invalid_Input";
		case Error_Other:
			return "Error_Other";
		default:
			return "Error_Other";
	}
	return "Error_Other";
}

static int operate_CheckCred(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	e_cred_t result = 0;

	char *username = dmjson_get_value((json_object *)value, 1, "Username");
	char *password = dmjson_get_value((json_object *)value, 1, "Password");

	result = check_user_cred(username, password);
	char *status = get_cred_result(result);
	add_list_parameter(ctx, dmstrdup("Status"), dmstrdup(status), DMT_TYPE[DMT_STRING], NULL);

	return 0;
}
static void update_group_participation(char *old_gr_name, char *new_gr_name, int op)
{
	if (old_gr_name == NULL)
		return;

	struct uci_section *s = NULL, *stmp = NULL;
	bbf_uci_foreach_sections_safe("users", "user", stmp, s) {
		char *pch = NULL, *spch = NULL, lbuf[1024] = {0};
		unsigned pos = 0;
		char *group_part = NULL;

		bbf_uci_get_value_by_section(s, "member_groups", &group_part);
		if (DM_STRLEN(group_part) == 0)
			continue;

		switch (op) {
		case NAME_CHANGE:
			if (new_gr_name == NULL)
				return;

			memset(lbuf, 0, sizeof(lbuf));
			for (pch = strtok_r(group_part, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
				if (strcmp(pch, old_gr_name) == 0) {
					pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", new_gr_name);
				} else {
					pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", pch);
				}
			}

			if (pos) {
				lbuf[pos - 1] = 0;
				bbf_uci_set_value_by_section(s, "member_groups", lbuf);
			}
			break;
		case DELETE:
			memset(lbuf, 0, sizeof(lbuf));
			for (pch = strtok_r(group_part, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
				if (strcmp(pch, old_gr_name) != 0) {
					pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", pch);
				}
			}

			if (pos)
				lbuf[pos - 1] = 0;

			bbf_uci_set_value_by_section(s, "member_groups", lbuf);
			break;
		}
	}
}

static void update_role_participation(char *old_name, char *new_name, int op, char *sect_type)
{
	if (old_name == NULL || sect_type == NULL)
		return;

	// Update roles of users or groups
	struct uci_section *s = NULL, *stmp = NULL;
	bbf_uci_foreach_sections_safe("users", sect_type, stmp, s) {
		char *pch = NULL, *spch = NULL, lbuf[1024] = {0};
		unsigned pos = 0;
		char *role_part = NULL;

		bbf_uci_get_value_by_section(s, "member_roles", &role_part);
		if (DM_STRLEN(role_part) == 0)
			continue;

		switch (op) {
		case NAME_CHANGE:
			if (new_name == NULL)
				return;

			memset(lbuf, 0, sizeof(lbuf));
			for (pch = strtok_r(role_part, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
				if (strcmp(pch, old_name) == 0) {
					pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", new_name);
				} else {
					pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", pch);
				}
			}

			if (pos) {
				lbuf[pos - 1] = 0;
				bbf_uci_set_value_by_section(s, "member_roles", lbuf);
			}
			break;
		case DELETE:
			memset(lbuf, 0, sizeof(lbuf));
			for (pch = strtok_r(role_part, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
				if (strcmp(pch, old_name) != 0) {
					pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", pch);
				}
			}

			if (pos)
				lbuf[pos - 1] = 0;

			bbf_uci_set_value_by_section(s, "member_roles", lbuf);
			break;
		}
	}
}

static void get_unique_role_id(char **id)
{
	unsigned int i;

	if (id == NULL)
		return;

	*id = NULL;
	for (i = 0; i < UINT_MAX; i++) {
		char str_id[30] = {0};
		snprintf(str_id, sizeof(str_id), "%u", i);
		if (get_dup_section_in_config_opt("users", "role", "role_id", str_id) == NULL) {
			*id = dmstrdup(str_id);
			break;
		}
	}

	return;
}

static bool user_exists(const char *user)
{
	struct uci_section *s = NULL, *stmp = NULL;
	uci_path_foreach_sections_safe(bbfdm, "dmmap_users", "user", stmp, s) {
		char *username = NULL;
		bbf_uci_get_value_by_section(s, "section_name", &username);
		if (DM_STRCMP(username, user) == 0)
			return true;
	}

	return false;
}

static void get_system_user_id(char *user_name, char **value)
{
	char cmd[256] = {0};

	if (DM_STRLEN(user_name) == 0) {
		*value = "";
		return;
	}

	snprintf(cmd, sizeof(cmd), "cat /etc/passwd | grep -r \"^%s:\" | cut -d: -f3", user_name);
	FILE *pp = popen(cmd, "r");
	if (pp != NULL) {
		char line[25] = {0};

		if (fgets(line, sizeof(line), pp) != NULL) {
			remove_new_line(line);
			*value = dmstrdup(line);
		} else {
			*value = "";
		}

		pclose(pp);
	} else {
		*value = "";
	}
}

static bool group_name_exists(const char *group_name)
{
	struct uci_section *s = NULL, *stmp = NULL;
	uci_path_foreach_sections_safe(bbfdm, "dmmap_users", "group", stmp, s) {
		char *name = NULL;
		bbf_uci_get_value_by_section(s, "section_name", &name);
		if (DM_STRCMP(name, group_name) == 0)
			return true;
	}

	return false;
}

static void get_system_group_id(char *group_name, char **value)
{
	char cmd[256] = {0};

	if (DM_STRLEN(group_name) == 0) {
		*value = "";
		return;
	}

	snprintf(cmd, sizeof(cmd), "cat /etc/group | grep -r \"^%s:\" | cut -d: -f3", group_name);
	FILE *pp = popen(cmd, "r");
	if (pp != NULL) {
		char line[25] = {0};

		if (fgets(line, sizeof(line), pp) != NULL) {
			remove_new_line(line);
			*value = dmstrdup(line);
		} else {
			*value = "";
		}

		pclose(pp);
	} else {
		*value = "";
	}
}

static bool static_entry(struct uci_section *s)
{
	char *value = NULL;
	bool val = false;

	if (s == NULL)
		return val;

	bbf_uci_get_value_by_section(s, "static", &value);
	if (DM_STRLEN(value) == 0)
		return val;
 
	bbf_convert_string_to_bool(value, &val);

	return val;
}

static void get_user_shell_name(char *username, char **shell)
{
	char cmd[128] = {0};

	if (username == NULL || shell == NULL)
		return;

	snprintf(cmd, sizeof(cmd), "cat /etc/passwd | grep -E \"^%s:\" | cut -d: -f7", username);
	FILE *pp = popen(cmd, "r");
	if (pp != NULL) {
		char line[2056] = {0};

		if (fgets(line, sizeof(line), pp) != NULL) {
			remove_new_line(line);
			if (DM_STRCMP(line, "/bin/false") != 0) {
				char *pos = strrchr(line, '/');
				*shell = pos ? dmstrdup(pos + 1) : dmstrdup(line);
			}
		}
		pclose(pp);
	}
}

static void get_perticipated_groups(char *username, char **groups)
{
	char cmd[128] = {0};

	if (username == NULL || groups == NULL)
		return;

	snprintf(cmd, sizeof(cmd), "groups %s", username);
	FILE *pp = popen(cmd, "r");
	if (pp != NULL) {
		char line[2056] = {0};

		if (fgets(line, sizeof(line), pp) != NULL) {
			remove_new_line(line);
			if (DM_STRNCMP(line, "groups:", 7) != 0) {
				*groups = dmstrdup(line);
			}
		}
		pclose(pp);
	}
}

static void sync_system_users_groups(FILE *pp, char *package, char *dmmap_package, char *section_type, bool user_cfg, struct list_head *dup_list)
{
	char line[65] = {0};

	if (pp == NULL || package == NULL || dmmap_package == NULL || section_type == NULL || dup_list == NULL)
		return;

	while (fgets(line, sizeof(line), pp) != NULL) {
		struct uci_section *dmmap_sect = NULL;
		char name[70] = {0};

		remove_new_line(line);
		if (!user_cfg) {
			snprintf(name, sizeof(name), "%s%s", GROUP_PREFIX, line);
		} else {
			snprintf(name, sizeof(name), "%s", line);
		}

		replace_special_char(name, '_');
		if (get_origin_section_from_config(package, section_type, name) != NULL) {
			continue;
		}

		if ((dmmap_sect = get_dup_section_in_dmmap(dmmap_package, section_type, name)) == NULL) {
			// add to dmmap if not present in users as well as dmmap_users file
			bbf_uci_add_section_bbfdm(dmmap_package, section_type, &dmmap_sect);
			bbf_uci_set_value_by_section(dmmap_sect, "section_name", name);
			bbf_uci_set_value_by_section(dmmap_sect, "static", "1");

			if (user_cfg) {
				char *shell = NULL, *group = NULL;
				get_user_shell_name(name, &shell);
				if (DM_STRLEN(shell) > 0)
					bbf_uci_set_value_by_section(dmmap_sect, "shell", shell);

				get_perticipated_groups(line, &group);
				if (DM_STRLEN(group) > 0) {
					char *pch = NULL, *spch = NULL, lbuf[1024] = {0};
					unsigned pos = 0;
					for (pch = strtok_r(group, " ", &spch); pch != NULL; pch = strtok_r(NULL, " ", &spch)) {
						if (DM_STRLEN(pch) < (1024 - strlen(lbuf) - strlen(GROUP_PREFIX))) {
							replace_special_char(pch, '_');
							pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s%s,", GROUP_PREFIX, pch);
						}
					}

					if (pos)
						lbuf[pos - 1] = 0;

					bbf_uci_set_value_by_section(dmmap_sect, "member_groups", lbuf);
				}
			}
		}

		/*
		 * Add dmmap sections to the list
		 */
		add_dmmap_config_dup_list(dup_list, NULL, dmmap_sect);
	}
}

static void synchronize_users_config_sections_with_dmmap(char *package, char *section_type, char *dmmap_package, struct list_head *dup_list, bool bootstrap)
{
	struct uci_section *s, *stmp;

	if (bootstrap == true) {
		char cmd[256] = {0};
		bool user_cfg;

		/* first copy the static users or groups to dmmap */
		if (DM_STRCMP(section_type, "user") == 0) {
			snprintf(cmd, sizeof(cmd), "cat /etc/passwd | cut -d: -f1");
			user_cfg = true;
		} else {
			snprintf(cmd, sizeof(cmd), "cat /etc/group | cut -d: -f1");
			user_cfg = false;
		}

		FILE *pp = popen(cmd, "r");
		if (pp != NULL) {
			sync_system_users_groups(pp, package, dmmap_package, section_type, user_cfg, dup_list);
			pclose(pp);
		}
	} else {
		/* add sytem users or groups section in list */
		uci_path_foreach_sections_safe(bbfdm, dmmap_package, section_type, stmp, s) {
			char *val = NULL;
			bool static_sec = false;
			bbf_uci_get_value_by_section(s, "static", &val);
			bbf_convert_string_to_bool(val, &static_sec);

			if (static_sec)
				add_dmmap_config_dup_list(dup_list, NULL, s);
		}
	}

	bbf_uci_foreach_sections(package, section_type, s) {
		char *val = NULL;
		bool deleted_sec = false;

		/* Do not include deleted group or user */
		bbf_uci_get_value_by_section(s, "deleted", &val);
		bbf_convert_string_to_bool(val, &deleted_sec);

		if (deleted_sec) {
			continue;
		}

		/*
		 * create/update corresponding dmmap section that have same config_section link and using param_value_array
		 */
		struct uci_section *dmmap_sect = NULL;

		if ((dmmap_sect = get_dup_section_in_dmmap(dmmap_package, section_type, section_name(s))) == NULL) {
			bbf_uci_add_section_bbfdm(dmmap_package, section_type, &dmmap_sect);
			bbf_uci_set_value_by_section(dmmap_sect, "section_name", section_name(s));
			bbf_uci_set_value_by_section(dmmap_sect, "static", "0");
		}

		/*
		 * Add system and dmmap sections to the list
		 */
		add_dmmap_config_dup_list(dup_list, s, dmmap_sect);
	}

	/*
	 * Delete unused non-static dmmap user/group sections
	 */
	uci_path_foreach_sections_safe(bbfdm, dmmap_package, section_type, stmp, s) {
		char *sec = NULL, *val = NULL;
		bool static_sec = false;

		bbf_uci_get_value_by_section(s, "section_name", &sec);
		bbf_uci_get_value_by_section(s, "static", &val);
		bbf_convert_string_to_bool(val, &static_sec);
		if (!static_sec && get_origin_section_from_config(package, section_type, sec) == NULL) {
			bbf_uci_delete_section_by_section(s, NULL, NULL);
		}
	}
}

/*************************************************************
* ENTRY METHOD
**************************************************************/
/*#Device.Users.User.{i}.!UCI:users/user/dmmap_users*/
static int browseUserInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	char *inst = NULL;
	struct dmmap_dup *p = NULL;
	LIST_HEAD(dup_list);
	bool bootstrap = false;

	char *filepath = "/var/run/user_bootstrap";
	if (file_exists(filepath)) {
		bootstrap = true;
		remove(filepath);
	}

	synchronize_users_config_sections_with_dmmap("users", "user", "dmmap_users", &dup_list, bootstrap);
	list_for_each_entry(p, &dup_list, list) {

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "user_instance", "user_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int browseGroupInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	char *inst = NULL;
	struct dmmap_dup *p = NULL;
	LIST_HEAD(dup_list);
	bool bootstrap = false;

	char *filepath = "/var/run/group_bootstrap";
	if (file_exists(filepath)) {
		bootstrap = true;
		remove(filepath);
	}

	synchronize_users_config_sections_with_dmmap("users", "group", "dmmap_users", &dup_list, bootstrap);
	list_for_each_entry(p, &dup_list, list) {

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "group_instance", "group_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int browseShellInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	char inst[30] = {0};
	int index = 1;
	struct uci_section *s, *stmp;

	uci_foreach_sections_safe("users", "shell", stmp, s) {
		snprintf(inst, sizeof(inst), "%d", index);
		index = index + 1;

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)s, inst) == DM_STOP)
			break;
	}

	return 0;
}

static int browseRoleInst(struct dmctx *dmctx, DMNODE *parent_node, void *prev_data, char *prev_instance)
{
	struct dmmap_dup *p = NULL;
	char *inst = NULL;
	LIST_HEAD(dup_list);

	synchronize_specific_config_sections_with_dmmap("users", "role", "dmmap_users", &dup_list);
	list_for_each_entry(p, &dup_list, list) {

		inst = handle_instance(dmctx, parent_node, p->dmmap_section, "role_instance", "role_alias");

		if (DM_LINK_INST_OBJ(dmctx, parent_node, (void *)p, inst) == DM_STOP)
			break;
	}
	free_dmmap_config_dup_list(&dup_list);
	return 0;
}

static int get_linker_group(char *refparam, struct dmctx *dmctx, void *data, char *instance, char **linker)
{
	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", linker);
	return 0;
}

static int get_linker_shell(char *refparam, struct dmctx *dmctx, void *data, char *instance, char **linker)
{
	bbf_uci_get_value_by_section((struct uci_section *)data, "name", linker);
	return 0;
}

static int get_linker_role(char *refparam, struct dmctx *dmctx, void *data, char *instance, char **linker)
{
	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "name", linker);
	return 0;
}

/*************************************************************
* ADD & DEL OBJ
**************************************************************/
static int add_users_user(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_user = NULL;
	char sec_name[40];

	snprintf(sec_name, sizeof(sec_name), "user_%s", *instance);

	if (bbf_uci_add_section("users", "user", &s) == 0) {
		bbf_uci_rename_section(s, sec_name);
		bbf_uci_set_value_by_section(s, "enabled", "0");
		bbf_uci_set_value_by_section(s, "remote_access", "0");
		bbf_uci_add_section_bbfdm("dmmap_users", "user", &dmmap_user);
		bbf_uci_set_value_by_section(dmmap_user, "section_name", sec_name);
		bbf_uci_set_value_by_section(dmmap_user, "user_instance", *instance);
		bbf_uci_set_value_by_section(dmmap_user, "static", "0");
	}

	return 0;
}

static int add_users_group(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_group = NULL;
	char sec_name[40];

	snprintf(sec_name, sizeof(sec_name), "%sgroup_%s", GROUP_PREFIX, *instance);

	if (bbf_uci_add_section("users", "group", &s) == 0) {
		bbf_uci_rename_section(s, sec_name);
		bbf_uci_set_value_by_section(s, "enabled", "0");
		bbf_uci_add_section_bbfdm("dmmap_users", "group", &dmmap_group);
		bbf_uci_set_value_by_section(dmmap_group, "section_name", sec_name);
		bbf_uci_set_value_by_section(dmmap_group, "group_instance", *instance);
		bbf_uci_set_value_by_section(dmmap_group, "static", "0");
	}

	return 0;
}

static int add_users_role(char *refparam, struct dmctx *ctx, void *data, char **instance)
{
	struct uci_section *s = NULL, *dmmap_role = NULL;
	char sec_name[40];
	char *role_id = NULL;

	snprintf(sec_name, sizeof(sec_name), "role_%s", *instance);
	get_unique_role_id(&role_id);
	if (role_id == NULL)
		return -1;

	if (bbf_uci_add_section("users", "role", &s) == 0) {
		char role_name[65] = {0};

		bbf_uci_rename_section(s, sec_name);
		bbf_uci_set_value_by_section(s, "enabled", "0");
		bbf_uci_set_value_by_section(s, "role_id", role_id);
		snprintf(role_name, sizeof(role_name), "Default_%s", *instance);
		bbf_uci_set_value_by_section(s, "name", role_name);

		bbf_uci_add_section_bbfdm("dmmap_users", "role", &dmmap_role);
		bbf_uci_set_value_by_section(dmmap_role, "section_name", sec_name);
		bbf_uci_set_value_by_section(dmmap_role, "role_instance", *instance);
	}

	return 0;
}

static int delete_users_user(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;

	switch (del_action) {
	case DEL_INST:
		if (!static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "deleted", "1");
			bbf_uci_delete_section_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
		}
		break;
	case DEL_ALL:
		bbf_uci_foreach_sections_safe("users", "user", stmp, s) {
			struct uci_section *dmmap_section = NULL;
			get_dmmap_section_of_config_section("dmmap_users", "user", section_name(s), &dmmap_section);
			if (!static_entry(dmmap_section)) {
				bbf_uci_set_value_by_section(s, "deleted", "1");
				bbf_uci_delete_section_by_section(dmmap_section, NULL, NULL);
			}
		}
		break;
	}
	return 0;
}

static int delete_users_group(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;
	char *gr_name = NULL;

	switch (del_action) {
	case DEL_INST:
		if (!static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", &gr_name);
			update_group_participation(gr_name, NULL, DELETE);
			bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "deleted", "1");
			bbf_uci_delete_section_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
		}
		break;
	case DEL_ALL:
		bbf_uci_foreach_sections_safe("users", "group", stmp, s) {
			struct uci_section *dmmap_section = NULL;
			get_dmmap_section_of_config_section("dmmap_users", "group", section_name(s), &dmmap_section);
			if (!static_entry(dmmap_section)) {
				update_group_participation(section_name(s), NULL, DELETE);
				bbf_uci_set_value_by_section(s, "deleted", "1");
				bbf_uci_delete_section_by_section(dmmap_section, NULL, NULL);
			}
		}
		break;
	}
	return 0;
}

static int delete_users_role(char *refparam, struct dmctx *ctx, void *data, char *instance, unsigned char del_action)
{
	struct uci_section *s = NULL, *stmp = NULL;
	char *role_name = NULL;

	switch (del_action) {
	case DEL_INST:
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "name", &role_name);
		update_role_participation(role_name, NULL, DELETE, "user");
		update_role_participation(role_name, NULL, DELETE, "group");
		bbf_uci_delete_section_by_section(((struct dmmap_dup *)data)->config_section, NULL, NULL);
		bbf_uci_delete_section_by_section(((struct dmmap_dup *)data)->dmmap_section, NULL, NULL);
		break;
	case DEL_ALL:
		bbf_uci_foreach_sections_safe("users", "role", stmp, s) {
			struct uci_section *dmmap_section = NULL;
			get_dmmap_section_of_config_section("dmmap_users", "role", section_name(s), &dmmap_section);
			bbf_uci_get_value_by_section(s, "name", &role_name);
			update_role_participation(role_name, NULL, DELETE, "user");
			update_role_participation(role_name, NULL, DELETE, "group");
			bbf_uci_delete_section_by_section(s, NULL, NULL);
			bbf_uci_delete_section_by_section(dmmap_section, NULL, NULL);
		}
		break;
	}
	return 0;
}

/*************************************************************
* GET & SET PARAM
**************************************************************/
static int get_users_user_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = bbf_get_number_of_entries(ctx, data, instance, browseUserInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_users_group_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = bbf_get_number_of_entries(ctx, data, instance, browseGroupInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_users_shell_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = bbf_get_number_of_entries(ctx, data, instance, browseShellInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_users_role_number_of_entries(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	int cnt = bbf_get_number_of_entries(ctx, data, instance, browseRoleInst);
	dmasprintf(value, "%d", cnt);
	return 0;
}

static int get_user_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "user_alias", instance, value);
}

static int get_group_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "group_alias", instance, value);
}

static int get_shell_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, (struct uci_section *)data, "shell_alias", instance, value);
}

static int get_role_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	return bbf_get_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "role_alias", instance, value);
}

static int get_shell_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = bbf_uci_get_value_by_section_fallback_def((struct uci_section *)data, "enabled", "1");
	return 0;
}

static int get_param_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (static_entry(((struct dmmap_dup *)data)->dmmap_section))
		*value = "1";
	else
		*value = bbf_uci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "enabled", "0");

	return 0;
}

static int get_user_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *sec_name = NULL;

	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", &sec_name);
	if (sec_name != NULL) {
		get_system_user_id(sec_name, value);
	} else {
		*value = "";
	}

	return 0;
}

static int get_group_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *sec_name = NULL, *group_name = NULL;

	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", &sec_name);
	if (sec_name != NULL) {
		if (DM_STRNCMP(sec_name, GROUP_PREFIX, strlen(GROUP_PREFIX)) == 0)
			group_name = dmstrdup(sec_name + strlen(GROUP_PREFIX));
		else
			group_name = dmstrdup(sec_name);

		get_system_group_id(group_name, value);
	} else {
		*value = "";
	}

	return 0;
}

static int get_role_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "role_id", value);
	return 0;
}


static int get_param_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *sec = NULL;
	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", &sec);

	if (DM_STRLEN(sec) != 0) {
		struct uci_section *s = ((struct dmmap_dup *)data)->dmmap_section;
		char *type = section_type(s);
		if (DM_STRCMP(type, "group") == 0) {
			*value = dmstrdup(sec+strlen(GROUP_PREFIX));
		} else {
			*value = dmstrdup(sec);
		}
	} else {
		*value = "";
	}

	return 0;
}

static int get_shell_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	bbf_uci_get_value_by_section((struct uci_section *)data, "name", value);
	return 0;
}

static int get_role_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "name", value);
	return 0;
}

static int get_user_password(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = "";
	return 0;
}

static int get_user_remote_accessable(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (static_entry(((struct dmmap_dup *)data)->dmmap_section))
		*value = "0";
	else
		*value = bbf_uci_get_value_by_section_fallback_def(((struct dmmap_dup *)data)->config_section, "remote_access", "1");

	return 0;
}

static int get_user_group_participation(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *groups = NULL;

	if (static_entry(((struct dmmap_dup *)data)->dmmap_section))
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "member_groups", &groups);
	else
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "member_groups", &groups);

	if (groups != NULL) {
		char *pch = NULL, *spch = NULL, lbuf[1024] = {0};
		unsigned pos = 0;

		lbuf[0] = 0;
		for (pch = strtok_r(groups, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			adm_entry_get_linker_param(ctx, "Device.Users.Group.", pch, value);
			if (DM_STRLEN(*value) > 0 && DM_STRLEN(*value) < (1024 - strlen(lbuf)))
				pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", *value);
		}

		if (pos)
			lbuf[pos - 1] = 0;

		*value = dmstrdup(lbuf);
	} else {
		*value = "";
	}

	return 0;
}

static int get_role_participation(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *roles = NULL;

	bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "member_roles", &roles);

	if (roles != NULL) {
		char *pch = NULL, *spch = NULL, lbuf[1024] = {0};
		unsigned pos = 0;

		lbuf[0] = 0;
		for (pch = strtok_r(roles, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			adm_entry_get_linker_param(ctx, "Device.Users.Role.", pch, value);
			if (DM_STRLEN(*value) > 0 && DM_STRLEN(*value) < (1024 - strlen(lbuf)))
				pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s,", *value);
		}

		if (pos)
			lbuf[pos - 1] = 0;

		*value = dmstrdup(lbuf);
	} else {
		*value = "";
	}

	return 0;
}

static int get_user_static_user(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
		*value = "1";
	} else {
		*value = "0";
	}

	return 0;
}

static int get_static_group(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
		*value = "1";
	} else {
		*value = "0";
	}

	return 0;
}

static int get_static_role(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	*value = "0";
	return 0;
}

static int get_user_language(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	if (static_entry(((struct dmmap_dup *)data)->dmmap_section))
		*value = "";
	else
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "language", value);

	return 0;
}

static int get_user_shell(char *refparam, struct dmctx *ctx, void *data, char *instance, char **value)
{
	char *shell = NULL;
	if (static_entry(((struct dmmap_dup *)data)->dmmap_section))
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "shell", &shell);
	else
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "shell", &shell);

	if (shell != NULL) {
		char *shell_path = NULL;
		adm_entry_get_linker_param(ctx, "Device.Users.SupportedShell.", shell, &shell_path);
		if (DM_STRLEN(shell_path) != 0)
			*value = dmstrdup(shell_path);
		else
			*value = "";
	} else {
		*value = "";
	}

	return 0;
}

static int set_user_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "user_alias", instance, value);
}

static int set_group_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "group_alias", instance, value);
}

static int set_shell_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, (struct uci_section *)data, "shell_alias", instance, value);
}

static int set_role_alias(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	return bbf_set_alias(ctx, ((struct dmmap_dup *)data)->dmmap_section, "role_alias", instance, value);
}

static int set_shell_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_boolean(ctx, value))
			return FAULT_9007;
		break;
	case VALUESET:
		bbf_convert_string_to_bool(value, &b);
		bbf_uci_set_value_by_section((struct uci_section *)data, "enabled", b ? "1" : "0");
		break;
	}
	return 0;
}

static int set_param_enable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;

	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_boolean(ctx, value))
			return FAULT_9007;
		break;
	case VALUESET:
		bbf_convert_string_to_bool(value, &b);
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "enabled", b ? "1" : "0");

		break;
	}
	return 0;
}

static int set_user_username(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char name[70] = {0};
	char *curr_name = NULL;

	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
			return FAULT_9007;

		// Check if the value is empty
		if (*value == '\0')
			return FAULT_9007;

		// Check if user already exist
		snprintf(name, sizeof(name), "%s", value);
		replace_special_char(name, '_');

		// Check if same as current username
		curr_name = section_name(((struct dmmap_dup *)data)->config_section);
		if (DM_STRCMP(curr_name, name) == 0)
			break;

		if (user_exists(name))
			return FAULT_9001;
		break;
	case VALUESET:
		snprintf(name, sizeof(name), "%s", value);
		replace_special_char(name, '_');

		// Check if same as current username
		curr_name = section_name(((struct dmmap_dup *)data)->config_section);
		if (DM_STRCMP(curr_name, name) == 0)
			break;

		// Update dmmap_users file
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", name);

		// Update users config
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "old_name",
						section_name(((struct dmmap_dup *)data)->config_section));
		bbf_uci_rename_section(((struct dmmap_dup *)data)->config_section, name);
		break;
	}
	return 0;
}

static int set_group_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *gr_name = NULL;
	char name[70] = {0};

	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
			return FAULT_9007;

		// Check if the value is empty
		if (*value == '\0')
			return FAULT_9007;

		snprintf(name, sizeof(name), "%s%s", GROUP_PREFIX, value);
		replace_special_char(name, '_');

		// Check if same as current groupname
		gr_name = section_name(((struct dmmap_dup *)data)->config_section);
		if (DM_STRCMP(gr_name, name) == 0)
			break;

		if (group_name_exists(name))
			return FAULT_9001;

		break;
	case VALUESET:
		gr_name = section_name(((struct dmmap_dup *)data)->config_section);
		snprintf(name, sizeof(name), "%s%s", GROUP_PREFIX, value);
		replace_special_char(name, '_');
		if (DM_STRCMP(gr_name, name) == 0)
			break;

		update_group_participation(gr_name, name, NAME_CHANGE);

		// Update dmmap_users file
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->dmmap_section, "section_name", name);

		// Update users config
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "old_name", gr_name);
		bbf_uci_rename_section(((struct dmmap_dup *)data)->config_section, name);

		break;
	}

	return 0;
}

static int set_role_name(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *role_name = NULL;

	switch (action) {
	case VALUECHECK:
		if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
			return FAULT_9007;

		// Check if the value is empty
		if (*value == '\0')
			return FAULT_9007;

		// Check if same as current role name
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "name", &role_name);
		if (DM_STRCMP(role_name, value) == 0)
			break;

		// Check if role name already exists
		if (get_dup_section_in_config_opt("users", "role", "name", value) != NULL)
			return FAULT_9001;

		break;
	case VALUESET:
		// get current role name
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "name", &role_name);
		if (DM_STRCMP(role_name, value) == 0)
			break;

		update_role_participation(role_name, value, NAME_CHANGE, "user");
		update_role_participation(role_name, value, NAME_CHANGE, "group");

		// Update users config
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "name", value);

		break;
	}

	return 0;
}

static int set_role_id(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *role_id = NULL;

	switch (action) {
	case VALUECHECK:
		// validate the value
		if (bbfdm_validate_unsignedInt(ctx, value, RANGE_ARGS{{NULL,NULL}}, 1))
			return FAULT_9007;

		// Check if same as current role id
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "role_id", &role_id);
		if (DM_STRCMP(role_id, value) == 0)
			break;

		// Check if this id already exists
		if (get_dup_section_in_config_opt("users", "role", "role_id", value) != NULL)
			return FAULT_9001;

		break;
	case VALUESET:
		// get current role id
		bbf_uci_get_value_by_section(((struct dmmap_dup *)data)->config_section, "role_id", &role_id);
		if (DM_STRCMP(role_id, value) == 0)
			break;

		// Update users config
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "role_id", value);

		break;
	}

	return 0;
}

static int set_user_password(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string(ctx, value, -1, 64, NULL, NULL))
			return FAULT_9007;
		break;
	case VALUESET:
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "password", value);
		break;
	}
	return 0;
}

static int set_user_remote_accessable(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	bool b;
	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_boolean(ctx, value))
			return FAULT_9007;
		break;
	case VALUESET:
		bbf_convert_string_to_bool(value, &b);
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "remote_access", b ? "1" : "0");
		break;
	}
	return 0;
}

static int set_user_group_participation(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *pch = NULL, *spch = NULL, new_group[1024] = { 0, 0 };
	char *allowed_obj[] = {
		"Device.Users.Group.",
		NULL
	};

	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string_list(ctx, value, -1, -1, 1024, -1, -1, NULL, NULL))
			return FAULT_9007;

		char *val = dmstrdup(value);
		for (pch = strtok_r(val, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			if (dm_entry_validate_allowed_objects(ctx, pch, allowed_obj))
				return FAULT_9007;
		}

		break;
	case VALUESET:
		new_group[0] = 0;
		unsigned pos = 0;

		for (pch = strtok_r(value, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			char *linker = NULL;
			adm_entry_get_linker_value(ctx, pch, &linker);
			if (!linker || linker[0] == '\0')
				continue;

			pos += snprintf(&new_group[pos], sizeof(new_group) - pos, "%s,", linker); 
		}

		if (pos)
			new_group[pos - 1] = 0;

		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "member_groups", new_group);
		break;
	}

	return 0;
}

static int set_role_participation(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *pch = NULL, *spch = NULL, new_role[1024] = { 0, 0 };
	char *allowed_obj[] = {
		"Device.Users.Role.",
		NULL
	};

	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string_list(ctx, value, -1, -1, 1024, -1, -1, NULL, NULL))
			return FAULT_9007;

		char *val = dmstrdup(value);
		for (pch = strtok_r(val, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			if (dm_entry_validate_allowed_objects(ctx, pch, allowed_obj))
				return FAULT_9007;
		}

		break;
	case VALUESET:
		new_role[0] = 0;
		unsigned pos = 0;

		for (pch = strtok_r(value, ",", &spch); pch != NULL; pch = strtok_r(NULL, ",", &spch)) {
			char *linker = NULL;
			adm_entry_get_linker_value(ctx, pch, &linker);
			if (!linker || linker[0] == '\0')
				continue;

			pos += snprintf(&new_role[pos], sizeof(new_role) - pos, "%s,", linker);
		}

		if (pos)
			new_role[pos - 1] = 0;

		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "member_roles", new_role);
		break;
	}

	return 0;
}

static int set_user_language(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string(ctx, value, -1, 16, NULL, NULL))
			return FAULT_9007;
		break;
	case VALUESET:
		bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "language", value);
		break;
	}
	return 0;
}

static int set_user_shell(char *refparam, struct dmctx *ctx, void *data, char *instance, char *value, int action)
{
	char *linker = NULL;
	char *allowed_obj[] = {
		"Device.Users.SupportedShell.",
		NULL
	};

	switch (action) {
	case VALUECHECK:
		if (static_entry(((struct dmmap_dup *)data)->dmmap_section)) {
			return FAULT_9001;
		}

		if (bbfdm_validate_string(ctx, value, -1, -1, NULL, NULL))
			return FAULT_9007;

		if (dm_entry_validate_allowed_objects(ctx, value, allowed_obj))
			return FAULT_9007;
		break;
	case VALUESET:
		adm_entry_get_linker_value(ctx, value, &linker);
		if (DM_STRLEN(linker) > 0)
			bbf_uci_set_value_by_section(((struct dmmap_dup *)data)->config_section, "shell", linker);

		break;
	}

	return 0;
}

/**********************************************************************************************************************************
*                                            OBJ & LEAF DEFINITION
***********************************************************************************************************************************/
DM_MAP_OBJ tDynamicObj[] = {
/* parentObj, nextObject, parameter */
{"Device.", tUsersObj, NULL},
{0}
};

DMOBJ tUsersObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys*/
{"Users", &DMREAD, NULL, NULL, NULL, NULL, NULL, NULL, tDeviceUsersObj, tUsersParams, NULL, BBFDM_BOTH, NULL},
{0}
};

/* *** Device.Users. *** */
DMOBJ tDeviceUsersObj[] = {
/* OBJ, permission, addobj, delobj, checkdep, browseinstobj, nextdynamicobj, dynamicleaf, nextobj, leaf, linker, bbfdm_type, uniqueKeys, version*/
{"User", &DMWRITE, add_users_user, delete_users_user, NULL, browseUserInst, NULL, NULL, NULL, tUsersUserParams, NULL, BBFDM_BOTH, LIST_KEY{"Username", "Alias", "UserID", NULL}},
{"Group", &DMWRITE, add_users_group, delete_users_group, NULL, browseGroupInst, NULL, NULL, NULL, tUsersGroupParams, get_linker_group, BBFDM_BOTH, LIST_KEY{"Groupname", "Alias", "GroupID", NULL}},
{"SupportedShell", &DMREAD, NULL, NULL, NULL, browseShellInst, NULL, NULL, NULL, tUsersShellParams, get_linker_shell, BBFDM_BOTH, LIST_KEY{"Name", "Alias", NULL}},
{"Role", &DMWRITE, add_users_role, delete_users_role, NULL, browseRoleInst, NULL, NULL, NULL, tUsersRoleParams, get_linker_role, BBFDM_BOTH, LIST_KEY{"Alias", "RoleName", "RoleID", NULL}},
{0}
};

DMLEAF tUsersParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"UserNumberOfEntries", &DMREAD, DMT_UNINT, get_users_user_number_of_entries, NULL, BBFDM_BOTH},
{"GroupNumberOfEntries", &DMREAD, DMT_UNINT, get_users_group_number_of_entries, NULL, BBFDM_BOTH},
{"RoleNumberOfEntries", &DMREAD, DMT_UNINT, get_users_role_number_of_entries, NULL, BBFDM_BOTH},
{"SupportedShellNumberOfEntries", &DMREAD, DMT_UNINT, get_users_shell_number_of_entries, NULL, BBFDM_BOTH},
{"CheckCredentialsDiagnostic()", &DMASYNC, DMT_COMMAND, get_operate_args_CheckCred, operate_CheckCred, BBFDM_USP},
{0}
};

/* *** Device.Users.User.{i}. *** */
DMLEAF tUsersUserParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Alias", &DMWRITE, DMT_STRING, get_user_alias, set_user_alias, BBFDM_BOTH},
{"Enable", &DMWRITE, DMT_BOOL, get_param_enable, set_param_enable, BBFDM_BOTH},
{"UserID", &DMREAD, DMT_UNINT, get_user_id, NULL, BBFDM_BOTH},
{"Username", &DMWRITE, DMT_STRING, get_param_name, set_user_username, BBFDM_BOTH},
{"Password", &DMWRITE, DMT_STRING, get_user_password, set_user_password, BBFDM_BOTH},
{"RemoteAccessCapable", &DMWRITE, DMT_BOOL, get_user_remote_accessable, set_user_remote_accessable, BBFDM_BOTH},
{"GroupParticipation", &DMWRITE, DMT_STRING, get_user_group_participation, set_user_group_participation, BBFDM_BOTH},
{"RoleParticipation", &DMWRITE, DMT_STRING, get_role_participation, set_role_participation, BBFDM_BOTH},
{"StaticUser", &DMREAD, DMT_BOOL, get_user_static_user, NULL, BBFDM_BOTH},
{"Language", &DMWRITE, DMT_STRING, get_user_language, set_user_language, BBFDM_BOTH},
{"Shell", &DMWRITE, DMT_STRING, get_user_shell, set_user_shell, BBFDM_BOTH},
{0}
};

/* *** Device.Users.Group.{i}. *** */
DMLEAF tUsersGroupParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Alias", &DMWRITE, DMT_STRING, get_group_alias, set_group_alias, BBFDM_BOTH},
{"Enable", &DMWRITE, DMT_BOOL, get_param_enable, set_param_enable, BBFDM_BOTH},
{"GroupID", &DMREAD, DMT_UNINT, get_group_id, NULL, BBFDM_BOTH},
{"Groupname", &DMWRITE, DMT_STRING, get_param_name, set_group_name, BBFDM_BOTH},
{"RoleParticipation", &DMWRITE, DMT_STRING, get_role_participation, set_role_participation, BBFDM_BOTH},
{"StaticGroup", &DMREAD, DMT_BOOL, get_static_group, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.Users.SupportedShell.{i}. *** */
DMLEAF tUsersShellParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Alias", &DMWRITE, DMT_STRING, get_shell_alias, set_shell_alias, BBFDM_BOTH},
{"Enable", &DMWRITE, DMT_BOOL, get_shell_enable, set_shell_enable, BBFDM_BOTH},
{"Name", &DMREAD, DMT_STRING, get_shell_name, NULL, BBFDM_BOTH},
{0}
};

/* *** Device.Users.Role.{i}. *** */
DMLEAF tUsersRoleParams[] = {
/* PARAM, permission, type, getvalue, setvalue, bbfdm_type, version*/
{"Alias", &DMWRITE, DMT_STRING, get_role_alias, set_role_alias, BBFDM_BOTH},
{"Enable", &DMWRITE, DMT_BOOL, get_param_enable, set_param_enable, BBFDM_BOTH},
{"RoleID", &DMWRITE, DMT_UNINT, get_role_id, set_role_id, BBFDM_BOTH},
{"RoleName", &DMWRITE, DMT_STRING, get_role_name, set_role_name, BBFDM_BOTH},
{"StaticRole", &DMREAD, DMT_BOOL, get_static_role, NULL, BBFDM_BOTH},
{0}
};
